/*
 * Programmer:  Timothy E. Hoff
 * File Name:   acl.c
 * Written:     July 5, 1996
 * Modified:
 * Description: This module provides access to ACL.
 *
 * Copyright (c) 1996, Timothy E. Hoff
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <grp.h>
#include "sfw.h"
#include "input.h"
#include "sizes.h"
#include "acl.h"
#include "error.h"
 
#ifndef _LINT
static char rcsid[] = "@(#) $Id: acl.c,v 2.1 1996/07/10 18:54:04 t07de7 Exp $";
#endif

/*
 * Private Function Declarations
 */

void static init_rec(struct ace *ace);
int static setfield(struct ace *ace, char *attr);
int setdays(int days[], char *value);
int static settime(time_t *nsec, char *value);
int setdate(time_t *nsec, char *value);
void static bad_attribute(char *keyword);
int static isgrpmem(char *, char *);

struct ace *getace(char *login, char *file, char *owner)
{
    /*
     * Lookup an access control entry for login in file.
     * File must be owned by either root or owner.  Return
     * a pointer to the ace structure or NULL if a matching
     * record could not be found.
     */

    static struct ace ace;
    FILE *fp;
    char *label, *attr;

    if ((fp = openfile(file, owner)) == NULL) {
        fprintf(stderr, "\nERROR:  Could not open %s.\n", file);
        fprintf(stderr, "Contact your system administrator.\n");
        exit(EXIT_FAILURE);
    }

    while ((label = nextstanza(fp)) != NULL) {
        /*
         * The label may represent a login name or a group name.
         * Groups are indicated by enclosing <>.
         */
        
        if (strcmp(label, login) == 0 ||
            (*label == '<' && isgrpmem(label, login))) {
            
            init_rec(&ace);
            strncpy(ace.login, login, MAXLOGIN - 1);
            
            while ((attr = getattr(fp)) != NULL) {
                if (setfield(&ace, attr) != 0) {
                    close(fp);
                    return NULL;
                }
            }

            close(fp);
            return &ace;
        }
    }

    close(fp);
    return NULL;
}


void static init_rec(struct ace *ace)
{
    int i;
    
    memset(ace->login, '\0', MAXLOGIN);
    for (i=0; i<7; i++)
        ace->days_allowed[i] = 1;
    
    ace->start_time = (time_t) 0;
    ace->end_time = (time_t) LONG_MAX;
    ace->expiration = (time_t) LONG_MAX;

    return;
}

    
int static setfield(struct ace *ace, char *attr)
{
    /*
     * Set the struct ace field which corresponds to the keyword portion
     * of attr to value.
     */

    char *keyword, *value;
    char msg[BUFSIZE];
    
    if ((keyword = strtok(attr, "=")) != NULL) {
        value = strtok(NULL, "\0");
    } else {
        /*
         * This code should never execute because getattr() in login.c
         * is supposed to filter for bad data.
         */
        sprintf(msg, "Invalid attribute [%s] found in ACL.", attr);
        log_error(ESYNTAX, msg);
        fprintf(stderr, "\nERROR:  Unrecognized attribute in ACL.\n");
        fprintf(stderr, "Contact your system administrator.\n");
        return -1;
    }

    switch (*keyword) {
        case 'd':
            if (strncmp(keyword, "days", strlen("days")) == 0) {
                setdays(ace->days_allowed, value);
            } else {
                bad_attribute(keyword);
                return -1;
            }
            break;
            
        case 'e':
            if (strncmp(keyword, "end", strlen("end")) == 0) {
                settime(&ace->end_time, value);
            } else if (strncmp(keyword, "expires", strlen("expires")) == 0) {
                setdate(&ace->expiration, value);
            } else {
                bad_attribute(keyword);
                return -1;
            }
            break;
            
        case 's':
            if (strncmp(keyword, "start", strlen("start")) == 0) {
                settime(&ace->start_time, value);
            } else {
                bad_attribute(keyword);
                return -1;
            }
            break;
        
        default:
            bad_attribute(keyword);
            return -1;
    }

    return 0;
}


int setdays(int days[], char *value)
{
    /*
     * Set a flag corresponding to each day listed in value.
     * Return 0 or -1 if an error occurred.
     */

    char *ch;
    char word[32];
    int i;

    /*
     * The default is to allow access every day.  The
     * current flags must be cleared to explicitly set
     * the access restrictions.
     */
     
    for (i=0; i<7; i++)
        days[i] = 0;
        
    ch = strtok(value, " ,");

    do {
        strncpy(word, ch, 32);
        for (i=0; i<32; i++) {
            if (word[i] != '\0')
                word[i] = tolower(word[i]);
        }

        switch (word[0]) {
            case 's':                   /* Sunday or Saturday */
                if (word[1] == 'u')
                    days[SUN] = 1;
                else if (word[1] == 'a')
                    days[SAT] = 1;
                else
                    return -1;
                break;
                
            case 'm':                   /* Monday */
                if (word[1] == 'o' || word[1] == '\0')
                    days[MON] = 1;
                else
                    return -1;
                break;
                
            case 't':                   /* Tuesday or Thursday */
                if (word[1] == 'u' || word[1] == '\0')
                    days[TUE] = 1;
                else if (word[1] == 'h')
                    days[THU] = 1;
                else
                    return -1;
                break;
                
            case 'w':                   /* Wednesday */
                if (word[1] == 'e' || word[1] == '\0')
                    days[WED] = 1;
                else
                    return -1;
                break;
                
            case 'r':                   /* Thursday */
                if (word[1] == '\0')
                    days[THU] = 1;
                else
                    return -1;
                break;
                
            case 'f':                   /* Friday */
                if (word[1] == 'r' || word[1] == '\0')
                    days[FRI] = 1;
                else
                    return -1;
                break;
                
            default:
                return -1;
        }
    } while (ch = strtok(NULL, " ,"));

    return 0;
}

        
int static settime(time_t *nsec, char *value)
{
    /*
     * Convert an ASCII time representation (HH:MM [a.m.|p.m.]) into
     * the number of seconds since the epoch.  Use the current month,
     * day and year to complete the conversion.  Return 0 on success
     * or -1 on an error.
     */

    time_t today;
    struct tm *ptr;
    char *hr, *m, *ch;
    int hours, min;
    int military = 1;               /* 24 hour clock */

    if ((hr = strtok(value, ":")) != NULL) {
        m = hr + strlen(hr) + 1;
        ch = m;
        while (*ch != '\0') {
            if (! isdigit(*ch)) {
                military = 0;
                break;
            }
	    ch++;
        }

        if (military)
            hours = atoi(hr);
        else 
            hours = (atoi(hr) + (*ch == 'p' ? 12 : 0) % 24);
            
        min = atoi(m);

        if (hours < 0 || hours > 23)
            return -1;

        if (min < 0 || min > 59)
            return -1;
    } else {
        return -1;
    }
        

    time(&today);               /* get the current time */
    ptr = localtime(&today);    /* break seconds into components */

    ptr->tm_min = min;
    ptr->tm_hour = hours;
    ptr->tm_sec = 0;

    *nsec = mktime(ptr);        /* convert back into seconds */

    if (nsec > 0)
        return 0;
    else
        return -1;
}


int setdate(time_t *nsec, char *value)
{
    /*
     * Set the date and time using value.  Value must be a string
     * formatted as mmddHHMM[[cc]yy] (see the UNIX date command).
     */

    char mon[3], mday[3], hour[3], min[3], year[5];
    time_t today;
    struct tm *ptr;
    char *ch = value;

    if (strlen(value) == 8 || strlen(value) == 10 || strlen(value) == 12) {

        memset(mon, '\0', 3);
        strncpy(mon, ch, 2);
        ch += 2;
    
        memset(mday, '\0', 3);
        strncpy(mday, ch, 2);
        ch += 2;
    
        memset(hour, '\0', 3);
        strncpy(hour, ch, 2);
        ch += 2;
    
        memset(min, '\0', 3);
        strncpy(min, ch, 2);
        ch += 2;

        memset(year, '\0', 5);
        strncpy(year, ch, 4);

    } else {
        return -1;
    }
    
    (void) time(&today);
    ptr = localtime(&today);

    ptr->tm_mon = atoi(mon);
    ptr->tm_mday = atoi(mday);
    ptr->tm_hour = atoi(hour);
    ptr->tm_min = atoi(min);

    if (strlen(year) == 2)
        ptr->tm_year = atoi(year);
    else if (strlen(year) == 4)
        ptr->tm_year = (atoi(year) - 1900);

    *nsec = mktime(ptr);

    if (nsec > 0)
        return 0;
    else
        return -1;
}


void static bad_attribute(char *keyword)
{
    char msg[BUFSIZE];

    sprintf(msg, "ACL keyword [%s] is not defined.", keyword);
    log_error(ESYNTAX, msg);

    return;    
}


int static isgrpmem(char *group, char *login)
{
    /*
     * Return 1 if login is a member of group otherwise 0.
     */

    struct group *grp;
    char **members;
    char buf[MAXGROUP];
    char *ch;
    char msg[BUFSIZE];

    strncpy(buf, group, MAXGROUP);

    if ((ch = strrchr(buf, '>')) == NULL) {
        sprintf(msg, "Invalid group label [%s].", group);
	log_error(EGROUP, msg);
        return 0;
    } else {
        *ch = '\0';
	group = &buf[1];
    }

    if ((grp = getgrnam(group)) == NULL) {
        return 0;
    } else {
        members = grp->gr_mem;
    }

    while (*members != NULL) {
        if (strcmp(login, *members) == 0) {
            return 1;
        }

        members++;
    }

    return 0;
}        
