/*
 * Programmer:  Timothy E. Hoff
 * File Name:   input.c
 * Written:     July 5, 1996
 * Modified:
 * Description: This module provides all read access to config files
 *              and the ACLs.
 *
 * Copyright (c) 1996, Timothy E. Hoff
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#ifdef _AIX
#include <sys/access.h>
#endif
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include "error.h"
#include "sizes.h"
#include "util.h"
#include "input.h"

 
#ifndef _LINT
static char rcsid[] = "@(#) $Id: input.c,v 2.3 1996/07/22 20:42:34 t07de7 Exp t07de7 $";
#endif



FILE *openfile(char *pathname, char *owner)
{
    /*
     * Open the file specified by pathname for reading.  Before returning
     * a pointer to the file verify that it is owned by owner or by root
     * and that it is not writable except by owner.  If any of these
     * conditions are not met return NULL even if the open would succeed.
     * Otherwise return a pointer to the open FILE.
     */

    FILE *fp;
    struct stat st;
    char buf[BUFSIZE];
    uid_t owner_id;

    if (*owner) {
        if ((owner_id = user_to_id(owner)) == -1) {
            sprintf(buf, "User [%s] is not defined.", owner);
            log_error(EPASSWD, buf);
            return NULL;
        }
    } else {
        owner = 0;         /* NULL owner; default root */
    }

    if (stat(pathname, &st) == 0) {
        if (! (st.st_uid == 0 || st.st_uid == owner_id)) {
            sprintf(buf, "%s is not owned by root or %s", pathname, owner);
            log_error(EOWNER, buf);
            return NULL;
        } /* file ownership is valid */
    } else {
        sprintf(buf, "stat(%s) failed with errno %d", pathname, errno);
        log_error(ENOENT, buf);
        return NULL;
    }

    if (accessx(pathname, W_OK, ACC_OTHERS) == 0) {
        sprintf(buf, "%s is writable by more than owner", pathname);
        log_error(EMODE, buf);
        return NULL;
    } /* file modes are valid */

    if ((fp = fopen(pathname, "rb")) != NULL) {
        return fp;
    } else {
        sprintf(buf, "open(%s) failed with errno %d", pathname, errno);
        log_error(errno, buf);
        return NULL;
    }
}


char *nextstanza(FILE *fp)
{
    /*
     * This function returns the label of the next stanza found in the file
     * pointed to by fp.  If an error occurs of EOF is reached the function
     * returns NULL.
     */

    static char buf[MAXLINE];
    char *ch, *label;
    int inword;

    while (fgets(buf, MAXLINE, fp) != NULL) {
        ch = buf;
        inword = 0;
        
        while (*ch != NULL) {
            if (*ch == ':') {           /* label */
                if (inword) {
                    *ch = '\0';
                    return label;
                } else {
                    log_error(ESYNTAX, "nextstanza:  Unexpected operator ':'");
                    return NULL;
                }                
            } else if (*ch == '#') {    /* comment */
                break;
            } else if (*ch == '=') {    /* attribute */
                break;
            } else if (*ch == '\n') {
                if (inword) {
                    log_error(ESYNTAX, "Unexpected end-of-line.  Ignoring.");
                    break;
                } else {
                    break;              /* blank line */
                }
            } else if (isprint(*ch)) {
                if (! inword) {
                    label = ch;
                    inword = 1;
                }
                ch++;
            } else {
                ch++;
            }
        } /* *ch != NULL */
    } /* EOF */

#ifdef _DEBUG
    if (feof(fp))
        fprintf(stderr, "DEBUG:  nextstanza reached EOF.\n");
    else
        fprintf(stderr, "DEBUG:  nextstanza failed on a read error.\n");
#endif
    return NULL;
}


char *getattr(FILE *fp)
{
    /*
     * This function returns the next attribute (keyword=value) from
     * the file pointed to by fp.  The function returns NULL on EOF,
     * the end of a stanza or an error.
     */

    static char buf[MAXLINE];
    static char attr[MAXLINE];
    char msg[BUFSIZE];
    char *ch, *keyword, *value;
    int lval = 0, op = 0, rval = 0;
    fpos_t pos;

    if (fgetpos(fp, &pos) !=0) {
        sprintf(msg, "fgetpos failed with errno %d", errno);
        log_error(EIO, msg);
    } else {
        if (fgets(buf, MAXLINE, fp) == NULL)
            return NULL;
    }
    
    do {        
        ch = buf;
        while (*ch != NULL) {

            if (*ch == ':') {           
                if (! op) {                         /* next stanza */
                    if (fsetpos(fp, &pos) != 0) {
                        sprintf(msg, "fsetpos failed with errno %d", errno);
                        log_error(EIO, msg);
                        exit(EXIT_FAILURE);
                    } else {
                        return NULL;
                    }
                } else if (! rval) {
                    value = ch;
                    rval = 1;
                }
                ch++;
            } else if (*ch == '#') {                /* comment */
                if (rval) {
                    *ch = '\0';
                    format_attr(attr, keyword, value);
                    return attr;
                } else if (op) {
                    log_error(ESYNTAX, "Syntax error:  Expected rval");
                    return NULL;
                } else if (lval) {
                    log_error(ESYNTAX, "Syntax error:  Expected '=' or ':'");
                    return NULL;
                } else {
                    break;
                }
            } else if (*ch == '=') {    /* attribute */
                if (lval && ! op) {
                    *ch = '\0';
                    ch++;
                    op = 1;
                } else if (rval) {
                    ch++;               /* embedded = in value */
                } else {
                    if (! op) 
                        log_error(ESYNTAX, "Syntax error:  Expected lval");
                    else
                        log_error(ESYNTAX, "Syntax error:  Unexpected operator '='");
                    return NULL;
                }
            } else if (*ch == '\n') {
                if (rval) {
                    *ch = '\0';
                    format_attr(attr, keyword, value);
                    return attr;
                } else if (op) {
                    log_error(ESYNTAX, "Syntax:  Expected rval");
                    return NULL;
                } else if (lval) {
                    log_error(ESYNTAX, "Syntax:  Expected '=' or ':'");
                    return NULL;
                } else {
                    break;              /* blank line */
                }
            } else if (isprint(*ch)) {
                if (! lval) {
                    keyword = ch;
                    lval = 1;
                } else if (op && ! rval) {
                    value = ch;
                    rval = 1;
                }
                ch++;
            } else {
                ch++;
            }
        } /* *ch != NULL */
        if (fgetpos(fp, &pos) !=0) {
            sprintf(msg, "fgetpos failed with errno %d", errno);
            log_error(EIO, msg);
        }
    } while (fgets(buf, MAXLINE, fp) != NULL);

    return NULL;
}


void format_attr(char *attr, char *keyword, char *value)
{
    /*
     * Write the attribute in canonical form to attr.  keyword and value
     * should not contain trailing spaces.
     */

    char *ch;

    for (ch = keyword + strlen(keyword); ch > keyword; ch--) {
        if (isspace(*ch))
            *ch = '\0';
        else
            break;
    }

    for (ch = (value + (strlen(value) - 1)); ch > value; ch--) {
        if (isspace(*ch))
            *ch = '\0';
        else
            break;
    }

    sprintf(attr, "%s=%s", keyword, value);
    return;
}
